/*
 * Copyright (C) 2004, 2006 Apple Computer, Inc.  All rights reserved.
 * Copyright (C) 2010 Patrick Gansterer <paroga@paroga.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "ResourceHandle.h"

#include "DataURL.h"
#include "HTTPParsers.h"
#include "MIMETypeRegistry.h"
#include "NotImplemented.h"
#include "ResourceError.h"
#include "ResourceHandleClient.h"
#include "ResourceHandleInternal.h"
#include "SharedBuffer.h"
#include "Timer.h"

#include <wtf/MainThread.h>
#include <wtf/UnusedParam.h>
#include <wtf/text/CString.h>

#include <WebCore/GCController.h>

#include "WebCoreInstanceHandle.h"

#include <windows.h>
#include <wininet.h>
#include "FileSystem.h"
#include <psapi.h>

// Internal Error Code
// Already reserved -999 in WebFrame.cpp
#define ERR_CODE_GZIP_OPEN		-998
#define ERR_CODE_GZIP_INFLATE	-997
#define ERR_CODE_SYNC			-996

#define INTERNET_BUFFER_SIZE   32768

// When formal release, DEBUG Trace is disable for performance.
#define MYDEBUGMSG( str, ... )      ((void)0)
#define MYDEBUGERR( str, errcode )  ((void)0)
//#define RHDEBUGMSG    MYDEBUGMSG
//#define RHDEBUGERR    MYDEBUGERR
#define RHDEBUGMSG( str, ... )      ((void)0)
#define RHDEBUGERR( str, errcode )  ((void)0)

#define TRYCATCH 1

#define RETURN_STATE_OK  true
#define RETURN_STATE_NG  false

#ifdef GLOBAL_INETHANDLE
static HINTERNET g_internetHandle = 0;
static HINTERNET g_internetHandleSync = 0;
#else
static DWORD g_flgInit = 0;
#endif

/* for garbageCollect */
static volatile DWORD g_thisCreateCounter = 0;
static int g_iniMemCheckInterval = 80;
static int g_iniMaxMemoryLoad = 78;
static DWORD g_iniMaxWorkingSet = 1610612736;   // 1.5G   (1395864371 = 1.3G)

static DWORD g_flagsHttpUseProtocolCachePolicy =  0x2040FE80;   // No Use Cache Mechanism of WinInet (If u want to use Cache Mechanism, set '0x40c800'
static DWORD g_flagsHttpReloadIgnoringCacheData = 0x8440F000;
static DWORD g_flagsHttpReturnCacheDataElseLoad = 0x2040FE80;
static DWORD g_flagsHttpReturnCacheDataDontLoad = 0x2040FE80;

#define EXTERNAL_EXENAME "webkitinet.exe"
static char* g_ExternalExeFilePtr = NULL;
static char  g_ExternalExeFile[256];

namespace WebCore {

static inline HINTERNET createInternetHandle(const String& userAgent, bool asynchronous) {
#ifdef GLOBAL_INETHANDLE
    if (!g_internetHandle && !g_internetHandleSync) {
#else
    if (!g_flgInit) {
        g_flgInit = 1;
#endif
        g_iniMemCheckInterval = WebCore::memCheckInterval();
        g_iniMaxMemoryLoad = WebCore::maxMemoryLoad();
        g_iniMaxWorkingSet = WebCore::maxWorkingSet();
        g_flagsHttpUseProtocolCachePolicy = WebCore::httpUseProtocolCachePolicy();
        g_flagsHttpReloadIgnoringCacheData = WebCore::httpReloadIgnoringCacheData();
        g_flagsHttpReturnCacheDataElseLoad = WebCore::httpReturnCacheDataElseLoad();
        g_flagsHttpReturnCacheDataDontLoad = WebCore::httpReturnCacheDataDontLoad();
        if (WebCore::httpSynchronousMode() & 1) {
            sprintf_s(g_ExternalExeFile, 256, "%s%s", WebCore::getModuleFolderName(), EXTERNAL_EXENAME);
            if (GetFileAttributesA(g_ExternalExeFile) != 0xFFFFFFFF) {
                g_ExternalExeFilePtr = g_ExternalExeFile;
            }
        }
		DWORD iHttpConnectionCnt = (WebCore::httpSpecialSetting() & 0xFF000000) >> 24;
        if (iHttpConnectionCnt) {
			InternetSetOption(NULL, INTERNET_OPTION_MAX_CONNS_PER_SERVER, &iHttpConnectionCnt, sizeof(DWORD));	// 1st param should be NULL when INTERNET_OPTION_MAX_CONNS_PER
			InternetSetOption(NULL, INTERNET_OPTION_MAX_CONNS_PER_1_0_SERVER, &iHttpConnectionCnt, sizeof(DWORD));
		}
    }
    String userAgentString = userAgent;
    HINTERNET internetHandle = InternetOpenW(userAgentString.charactersWithNullTermination(), INTERNET_OPEN_TYPE_PRECONFIG, 0, 0, asynchronous ? INTERNET_FLAG_ASYNC : 0);

    if (internetHandle) {
        DWORD ms = WebCore::httpDefaultTimeOut();
        if (ms > 0) {
            InternetSetOption(internetHandle, INTERNET_OPTION_CONNECT_TIMEOUT, &ms, sizeof(ms));
            InternetSetOption(internetHandle, INTERNET_OPTION_CONTROL_RECEIVE_TIMEOUT, &ms, sizeof(ms));
            InternetSetOption(internetHandle, INTERNET_OPTION_CONTROL_SEND_TIMEOUT, &ms, sizeof(ms));
            InternetSetOption(internetHandle, INTERNET_OPTION_DATA_SEND_TIMEOUT, &ms, sizeof(ms));
            InternetSetOption(internetHandle, INTERNET_OPTION_DATA_RECEIVE_TIMEOUT, &ms, sizeof(ms));
        }
	    if (WebCore::gzipContentEncoding() & 4) {   // init@C3bit(4)w肳ꂽꍇ
	        BOOL bEnable = TRUE;
            BOOL sts = InternetSetOption(internetHandle, INTERNET_OPTION_HTTP_DECODING, &bEnable, sizeof(bEnable));
		    MYDEBUGMSG("InternetSetOption sts = %d\n", sts);
		}

        if (asynchronous)
            InternetSetStatusCallback(internetHandle, &ResourceHandle::internetStatusCallback);
    }
    return internetHandle;
}

static HINTERNET setInternetHandle(const String& userAgent, bool asynchronous) {
    static HINTERNET internetHandle = createInternetHandle(userAgent, asynchronous);
    return internetHandle;
}

static String queryHTTPHeader(HINTERNET requestHandle, DWORD infoLevel) {
    DWORD bufferSize = 0;
    HttpQueryInfoW(requestHandle, infoLevel, 0, &bufferSize, 0);

    Vector<UChar> characters(bufferSize / sizeof(UChar));

    if (!HttpQueryInfoW(requestHandle, infoLevel, characters.data(), &bufferSize, 0))
        return String();

    characters.removeLast(); // Remove NullTermination.
    return String::adopt(characters);
}

static DWORD getInetFlags(const ResourceRequest& request) {
    DWORD iflags = g_flagsHttpUseProtocolCachePolicy;   // load default
    switch (request.cachePolicy()) {
    case UseProtocolCachePolicy:
        RHDEBUGMSG("####### UseProtocolCachePolicy #######\n");
        break;
    case ReloadIgnoringCacheData:
        RHDEBUGMSG("####### ReloadIgnoringCacheData #######\n");
        iflags = g_flagsHttpReloadIgnoringCacheData;
        break;
    case ReturnCacheDataElseLoad:
        RHDEBUGMSG("####### ReturnCacheDataElseLoad #######\n");
        iflags = g_flagsHttpReturnCacheDataElseLoad;
        break;
    case ReturnCacheDataDontLoad:
        RHDEBUGMSG("####### ReturnCacheDataDontLoad #######\n");
        iflags = g_flagsHttpReturnCacheDataDontLoad;
        break;
    default:
        RHDEBUGMSG("####### default policy = 0x%x#######\n", request.cachePolicy());
    }

    if (request.url().protocolIs("https")) {
        iflags |= INTERNET_FLAG_SECURE;
    }
    return iflags;
}

static Vector<UChar> getRequestHeaders(const ResourceRequest& request) {
    Vector<UChar> httpHeaders;
    const HTTPHeaderMap& httpHeaderFields = request.httpHeaderFields();

    for (HTTPHeaderMap::const_iterator it = httpHeaderFields.begin(); it != httpHeaderFields.end(); ++it) {
        if (equalIgnoringCase(it->key, "Accept") || equalIgnoringCase(it->key, "Referer") || equalIgnoringCase(it->key, "User-Agent"))
            continue;
        if (WebCore::httpSpecialSetting() & 1) {
            // init@C1bitڂw肳ꂽꍇ̂݁AIf-Modifiedn𖳌
            // WebKitWebCacheLɁACachedResource.cppCachedResource::load֐
            // If-ModifiednNGXgoB̍ۂɏQo͖{L(ftHgł͖{͖)
            if (equalIgnoringCase(it->key, "Cache-Control") || equalIgnoringCase(it->key, "If-Modified-Since") || equalIgnoringCase(it->key, "If-None-Match"))
                continue;
        }

        httpHeaders.append(it->key.characters(), it->key.length());
        httpHeaders.append(':');
        httpHeaders.append(it->value.characters(), it->value.length());
        httpHeaders.append('\r');
        httpHeaders.append('\n');
    }

    if (WebCore::gzipContentEncoding() & 2) {   // init@C2bitڂw肳ꂽꍇ̂݁AϋɓIgzipv(option)
        String strKey = String("Accept-Encoding");
        String strVal = String("gzip,deflate");
        httpHeaders.append(strKey.characters(), strKey.length());
        httpHeaders.append(':');
        httpHeaders.append(strVal.characters(), strVal.length());
        httpHeaders.append('\r');
        httpHeaders.append('\n');
    }
    // Some web sites check "referer" header.
    // Maybe these codes are necessary.
    String httpReferrer = request.httpReferrer();
    if (!httpReferrer.isEmpty()) {
        String strKey2 = String("Referer");
        httpHeaders.append(strKey2.characters(), strKey2.length());
        httpHeaders.append(':');
        httpHeaders.append(httpReferrer.characters(), httpReferrer.length());
        httpHeaders.append('\r');
        httpHeaders.append('\n');
    }
    return httpHeaders;
}

class WebCoreSynchronousLoader : public ResourceHandleClient {
    WTF_MAKE_NONCOPYABLE(WebCoreSynchronousLoader);
public:
    WebCoreSynchronousLoader(ResourceError&, ResourceResponse&, Vector<char>&, const String& userAgent);
    ~WebCoreSynchronousLoader();

#ifdef GLOBAL_INETHANDLE
    HINTERNET internetHandle() const {
        return g_internetHandleSync;
    }
#else
    HINTERNET internetHandle() const {
        return m_internetHandle;
    }
#endif
    virtual void didReceiveResponse(ResourceHandle*, const ResourceResponse&);
    virtual void didReceiveData(ResourceHandle*, const char*, int, int encodedDataLength);
    virtual void didFinishLoading(ResourceHandle*, double /*finishTime*/);
    virtual void didFail(ResourceHandle*, const ResourceError&);

private:
    ResourceError& m_error;
    ResourceResponse& m_response;
    Vector<char>& m_data;
#ifndef GLOBAL_INETHANDLE
    HINTERNET m_internetHandle;
#endif
};

WebCoreSynchronousLoader::WebCoreSynchronousLoader(ResourceError& error, ResourceResponse& response, Vector<char>& data, const String& userAgent)
    : m_error(error)
    , m_response(response)
    , m_data(data)
#ifndef GLOBAL_INETHANDLE
    , m_internetHandle(createInternetHandle(userAgent, false))
#endif
{
    RHDEBUGMSG("start\n");
#ifdef GLOBAL_INETHANDLE
    if (!g_internetHandleSync) g_internetHandleSync = createInternetHandle(userAgent, false);
#endif
}

WebCoreSynchronousLoader::~WebCoreSynchronousLoader() {
    RHDEBUGMSG("start\n");
#ifndef GLOBAL_INETHANDLE
    InternetCloseHandle(m_internetHandle);
#endif
}

void WebCoreSynchronousLoader::didReceiveResponse(ResourceHandle*, const ResourceResponse& response) {
    m_response = response;
}

void WebCoreSynchronousLoader::didReceiveData(ResourceHandle*, const char* data, int length, int) {
    m_data.append(data, length);
}

void WebCoreSynchronousLoader::didFinishLoading(ResourceHandle*, double) {
}

void WebCoreSynchronousLoader::didFail(ResourceHandle*, const ResourceError& error) {
    m_error = error;
}

/* WebCore Global Function called by WebKitDLL.cpp(WebKitLib) */
void ResourceHandleInternal::closeInternetHandle() {
    RHDEBUGMSG("start\n");
#ifdef GLOBAL_INETHANDLE
    if (g_internetHandle) {
        InternetCloseHandle(g_internetHandle);
        g_internetHandle = 0;
    }
    if (g_internetHandleSync) {
        InternetCloseHandle(g_internetHandleSync);
        g_internetHandleSync = 0;
    }
#endif
}

ResourceHandleInternal::~ResourceHandleInternal() {
    RHDEBUGMSG("start\n");
    if (m_fileHandle) {
        CloseHandle(m_fileHandle);
        m_fileHandle = 0;
    }
    if (m_fileLoadTimer.isActive()) m_fileLoadTimer.stop();
    RHDEBUGMSG("end\n");
}

ResourceHandle::~ResourceHandle() {
    // RHDEBUGMSG("start this = 0x%x, firstRequest().url() = %s\n", this, firstRequest().url().string().latin1().data());
    clearHandle();
}

static void callOnRedirect(void* context) {
    RHDEBUGMSG("start context = 0x%x\n", context);
    ResourceHandle* handle = static_cast<ResourceHandle*>(context);
    handle->onRedirect();
}

static void callOnRequestStart(void* context) {
    RHDEBUGMSG("start context = 0x%x\n", context);
    ResourceHandle* handle = static_cast<ResourceHandle*>(context);
    handle->onRequestStart();
}

static void callOnRequestComplete(void* context) {
    RHDEBUGMSG("start context = 0x%x\n", context);
    ResourceHandle* handle = static_cast<ResourceHandle*>(context);
    handle->onRequestComplete();
}

static void callOnReadFile(void* context) {
    ResourceHandle* handle = static_cast<ResourceHandle*>(context);
    handle->onReadFile();
}

void ResourceHandle::internetStatusCallback(HINTERNET internetHandle, DWORD_PTR context, DWORD internetStatus,
        LPVOID statusInformation, DWORD statusInformationLength) {
#ifdef TRYCATCH
    try {
#endif
        ResourceHandle* handle = reinterpret_cast<ResourceHandle*>(context);
        RHDEBUGMSG("start this = 0x%x, ThreadID = 0x%x, internetStatus = %d\n", handle, GetCurrentThreadId(), internetStatus);

        switch (internetStatus) {
        case INTERNET_STATUS_HANDLE_CREATED:
            if (!handle->d->m_connectHandle) {
                if (!handle->d->m_iowaiting) {
                    RHDEBUGMSG("INTERNET_STATUS_HANDLE_CREATED m_connectHandle = 0x%x\n", handle->d->m_connectHandle);
                    return;
                }
                handle->d->m_iowaiting = false;
                INTERNET_ASYNC_RESULT *pRes = (INTERNET_ASYNC_RESULT *)statusInformation;
                handle->d->m_connectHandle = (HINTERNET)pRes->dwResult;
                RHDEBUGMSG("INTERNET_STATUS_HANDLE_CREATED set m_connectHandle = 0x%x\n", handle->d->m_connectHandle);
                callOnMainThread(callOnRequestStart, handle);
            } else if (!handle->d->m_requestHandle) {
                if (!handle->d->m_iowaiting) {
                    RHDEBUGMSG("INTERNET_STATUS_HANDLE_CREATED m_requestHandle = 0x%x\n", handle->d->m_requestHandle);
                    return;
                }
                handle->d->m_iowaiting = false;
                INTERNET_ASYNC_RESULT *pRes = (INTERNET_ASYNC_RESULT *)statusInformation;
                handle->d->m_requestHandle = (HINTERNET)pRes->dwResult;
                RHDEBUGMSG("INTERNET_STATUS_HANDLE_CREATED set m_requestHandle = 0x%x\n", handle->d->m_requestHandle);
                callOnMainThread(callOnRequestStart, handle);
            } else {
                MYDEBUGMSG("INTERNET_STATUS_HANDLE_CREATED other\n");
            }
            break;

        case INTERNET_STATUS_REDIRECT:	{
			//Multiple language support Fixed by Lupin NECSTHZ
			//̃XgO̒ 			
			//handle->d->m_redirectUrl = String(static_cast<UChar*>(statusInformation), statusInformationLength);
			UChar* str=static_cast<UChar*>(statusInformation);
			int len = WideCharToMultiByte( CP_ACP,0,str,-1,NULL,0,NULL,NULL ); 
			handle->d->m_redirectUrl = String(str, len);
			
            // Tsv
            //if (strncmp(handle->d->m_redirectUrl.latin1().data(), "https", 5) == 0) {
            //  handle->d->m_flags |= INTERNET_FLAG_SECURE;
            //} else {
            //  handle->d->m_flags &= (~INTERNET_FLAG_SECURE);
            //}
            RHDEBUGMSG("INTERNET_STATUS_REDIRECT start m_redirectUrl = %s, m_flags = 0x%x\n", handle->d->m_redirectUrl.latin1().data(), handle->d->m_flags);
            callOnMainThread(callOnRedirect, handle);
			}
            break;

        case INTERNET_STATUS_REQUEST_COMPLETE: {
                INTERNET_ASYNC_RESULT *pAsyncRes = (INTERNET_ASYNC_RESULT *)statusInformation;
                RHDEBUGMSG("INTERNET_STATUS_REQUEST_COMPLETE dwResult = %d, dwError = %d\n", pAsyncRes->dwResult, pAsyncRes->dwError);
                if (handle->d->m_connectHandle) {
                    if ((handle->d->m_dwErrorCode = pAsyncRes->dwError) != ERROR_INTERNET_OPERATION_CANCELLED) {
                        callOnMainThread(callOnRequestComplete, handle);
                        // ERROR_INTERNET_OPERATION_CANCELLEDƁAݎsonRequestCompleteWinInet֐̓G[IB
                    }
                } else {
                    RHDEBUGMSG("INTERNET_STATUS_REQUEST_COMPLETE already clear m_connectHandle\n", pAsyncRes->dwResult, pAsyncRes->dwError);
                }
            }
            break;

        case INTERNET_STATUS_REQUEST_SENT: {
                DWORD *lpBytesSent = (DWORD*)statusInformation;
                RHDEBUGMSG("INTERNET_STATUS_REQUEST_SENT lpBytesSent = %d\n", *lpBytesSent);
            }
            break;
        case INTERNET_STATUS_RESPONSE_RECEIVED: {
                DWORD *dwBytesReceived = (DWORD*)statusInformation;
                RHDEBUGMSG("INTERNET_STATUS_RESPONSE_RECEIVED dwBytesReceived = %d\n", *dwBytesReceived);
            }
            break;

        case INTERNET_STATUS_CONNECTION_CLOSED: {
                RHDEBUGMSG("INTERNET_STATUS_CONNECTION_CLOSED\n");
                //handle->d->m_requestHandle = NULL;
                //handle->d->m_connectHandle = NULL;
            }
            break;
        case INTERNET_STATUS_SENDING_REQUEST: { // 30
                RHDEBUGMSG("INTERNET_STATUS_SENDING_REQUEST\n");
            }
            break;
        case INTERNET_STATUS_RECEIVING_RESPONSE: {  // 40
                RHDEBUGMSG("INTERNET_STATUS_RECEIVING_RESPONSE\n");
            }
            break;
        case INTERNET_STATUS_CLOSING_CONNECTION: {  // 50
                RHDEBUGMSG("INTERNET_STATUS_CLOSING_CONNECTION\n");
            }
            break;
        case INTERNET_STATUS_HANDLE_CLOSING: {  // 70
                RHDEBUGMSG("INTERNET_STATUS_HANDLE_CLOSING\n");
                //if (IsBadReadPtr(handle, sizeof(ResourceHandle)) == 0) {
                //  if (handle->d) {
                //      if (IsBadReadPtr(&handle->d->m_dwHandleClosing, sizeof(DWORD)) == 0) {
                //          handle->d->m_dwHandleClosing++;
                //          RHDEBUGMSG("INTERNET_STATUS_HANDLE_CLOSING m_dwHandleClosing = %d\n", handle->d->m_dwHandleClosing);
                //      }
                //  }
                //}
            }
            break;
        case INTERNET_STATUS_COOKIE_SENT: { // 320
                RHDEBUGMSG("INTERNET_STATUS_COOKIE_SENT\n");
            }
            break;
        case INTERNET_STATUS_P3P_HEADER: {  // 325
                RHDEBUGMSG("INTERNET_STATUS_P3P_HEADER\n");
            }
            break;
        case INTERNET_STATUS_P3P_POLICYREF: { // 326
                RHDEBUGMSG("INTERNET_STATUS_P3P_POLICYREF\n");
            }
            break;
        default:
            RHDEBUGMSG("internetStatus = %d\n", internetStatus);
            break;
        }

#ifdef TRYCATCH
    } catch(...) {  // OLb`
        MYDEBUGMSG("internetStatusCallback exception!!!\n");
    }
#endif
}

void ResourceHandle::FileReadIOCompletionRoutine(DWORD dwErrorCode, DWORD dwNumberOfBytesTransfered, LPOVERLAPPED lpOverlapped) {
    RHDEBUGMSG("start ThreadID = 0x%x, dwErrorCode = 0x%x, dwNumberOfBytesTransfered = %d\n", GetCurrentThreadId(), dwErrorCode, dwNumberOfBytesTransfered);
    ResourceHandle* handle = reinterpret_cast<ResourceHandle*>(lpOverlapped->hEvent);
    handle->d->m_dwErrorCode = dwErrorCode;
    handle->d->m_dwNumberOfBytesTransfered = dwNumberOfBytesTransfered;
    callOnMainThread(callOnReadFile, handle);
}

// Dummy Entry for Rabbit
void ResourceHandle::setHostAllowsAnyHTTPSCertificate(const String&) {
    MYDEBUGMSG("start  **** notImplemented !!! ****\n");
}
// Dummy Entry for Rabbit
void ResourceHandle::setClientCertificate(const String& host, CFDataRef) {
    MYDEBUGMSG("start  **** notImplemented !!! ****\n");
}

void ResourceHandle::clearHandle() {
    RHDEBUGMSG("start this = 0x%x\n", this);
    if (d->m_requestHandle) {
        InternetCloseHandle(d->m_requestHandle);
        d->m_requestHandle = NULL;
    }
    if (d->m_connectHandle) {
        InternetCloseHandle(d->m_connectHandle);
        d->m_connectHandle = NULL;
    }
#ifndef GLOBAL_INETHANDLE
    if (d->m_internetHandle) {
        // 񓯊nh̏ꍇAUCloseƈȍ~InternetConnectWs邽߁A
        // CloseȂBWebKit.org̃IWi̎B
//      InternetCloseHandle(d->m_internetHandle);
        d->m_internetHandle = 0;
    }
#endif
	if (d->m_gzipInflateBuf) {
		inflateEnd(&d->m_zst);
		HeapFree(GetProcessHeap(),0,d->m_gzipInflateBuf);
		d->m_gzipInflateBuf = NULL;
	}

	if (m_requestCompleteEvent[0]){
		CloseHandle(m_requestCompleteEvent[0]);
		m_requestCompleteEvent[0] = NULL;
	}

	if (m_requestCompleteEvent[1]){
		CloseHandle(m_requestCompleteEvent[1]);
		m_requestCompleteEvent[1] = NULL;
	}
}

bool ResourceHandle::onRequestStart() {
    RHDEBUGMSG("start this = 0x%x, client() = 0x%x, ThreadID = 0x%x, url = %s\n", this, client(), GetCurrentThreadId(), firstRequest().url().string().latin1().data());
    if (!d->m_requestHandle) {
        DWORD rqFlagsHttpAccess = d->m_flags;

		if (firstRequest().httpBody()) {
             rqFlagsHttpAccess |= INTERNET_FLAG_RELOAD | INTERNET_FLAG_NO_CACHE_WRITE ;
		}

        String urlStr = firstRequest().url().path();
        String urlQuery = firstRequest().url().query();
        if (!urlQuery.isEmpty()) {
            urlStr.append('?');
            urlStr.append(urlQuery);
        }
        String httpMethod = firstRequest().httpMethod();
        LPCWSTR httpAccept[] = { L"*/*", 0 };

        RHDEBUGMSG("start HttpOpenRequestW Request_Type = %s, url = %s, WinInet_Flags = 0x%x\n", httpMethod.latin1().data(), urlStr.latin1().data(), rqFlagsHttpAccess);
        // 5Ԗڂ̈httpReferrer.charactersWithNullTermination() ݒ肷ƁAhttps -> http redirect ɂăG[8AʐMsłȂȂ̂ŒӁI
        d->m_requestHandle = HttpOpenRequestW(d->m_connectHandle, httpMethod.charactersWithNullTermination(), urlStr.charactersWithNullTermination(),
                                              0, 0, httpAccept, rqFlagsHttpAccess, d->m_loadSynchronously ? 0 : reinterpret_cast<DWORD_PTR>(this));
        if (d->m_requestHandle) {
            RHDEBUGMSG("success HttpOpenRequestW m_requestHandle = 0x%x\n", d->m_requestHandle);
        } else {
            DWORD dwError = GetLastError();
            if (dwError == ERROR_IO_PENDING) {
                RHDEBUGMSG("ERROR_IO_PENDING HttpOpenRequestW\n");
                d->m_iowaiting = true;
                return RETURN_STATE_OK;
            } else {
                clearHandle();
                MYDEBUGERR("HttpOpenRequestW didFail", dwError);
                if (ResourceHandleClient* resourceHandleClient = client())
                    resourceHandleClient->didFail(this, ResourceError(firstRequest().url().host(), dwError, firstRequest().url().string(), ""));
                return RETURN_STATE_NG;
            }
        }
    }

	if (firstRequest().httpBody()) {
		d->m_bytesRemainingToWrite = firstRequest().httpBody()->getSize();
		if ( d->m_bytesRemainingToWrite > 0) {
			d->m_hasRemainingToWrite = true;
		}
        
		m_requestCompleteEvent[0] = CreateEvent(NULL, FALSE, FALSE, NULL);
		m_requestCompleteEvent[1] = CreateEvent(NULL, FALSE, FALSE, NULL);
	}	

    Vector<UChar> httpHeaders = getRequestHeaders(firstRequest());
    INTERNET_BUFFERSW internetBuffers;
    ZeroMemory(&internetBuffers, sizeof(internetBuffers));
    internetBuffers.dwStructSize = sizeof(internetBuffers);
    internetBuffers.lpcszHeader = String(httpHeaders.data(), httpHeaders.size()).charactersWithNullTermination();
    internetBuffers.dwHeadersLength = httpHeaders.size();		// should keep dwHeadersTotal 0 
    internetBuffers.dwBufferTotal = d->m_bytesRemainingToWrite; // should keep dwBufferLength 0 
	//if (internetBuffers.dwBufferTotal > 1024) {
	//	d->m_uploadcnt = (internetBuffers.dwBufferTotal / 1024) + 1;
	//} else {
	//	d->m_uploadcnt = 1;
	//}
 //   MYDEBUGMSG("internetBuffers.dwBufferTotal = %d, d->m_uploadcnt = %d\n", internetBuffers.dwBufferTotal, d->m_uploadcnt);
#if 0
    char* buf2 = (char*)malloc(internetBuffers.dwHeadersLength + 1);
    memset(buf2, 0, internetBuffers.dwHeadersLength + 1);
    WideCharToMultiByte(CP_ACP, WC_NO_BEST_FIT_CHARS, internetBuffers.lpcszHeader, -1, buf2, internetBuffers.dwHeadersLength, NULL, NULL);
    MYDEBUGMSG("start HttpSendRequestExW sjislen = %d, httpHeaders = \n----\n%s\n----\n", strlen(buf2), buf2);
    free(buf2);
#endif
    DWORD dwError = 0;
    BOOL sts = HttpSendRequestExW(d->m_requestHandle, &internetBuffers, 0, 0, d->m_loadSynchronously ? 0 : reinterpret_cast<DWORD_PTR>(this));
    if (!sts) {
        dwError = GetLastError();
        if (dwError != ERROR_IO_PENDING) {
            MYDEBUGERR("HttpSendRequestEx", dwError);
            MYDEBUGMSG("url = %s\n", firstRequest().url().string().latin1().data());
            if ( dwError == ERROR_INTERNET_CLIENT_AUTH_CERT_NEEDED ) {
                // Return ERROR_SUCCESS regardless of clicking on OK or Cancel
                sts = true;
                if (InternetErrorDlg( GetDesktopWindow(),
                                      d->m_requestHandle,
                                      ERROR_INTERNET_CLIENT_AUTH_CERT_NEEDED,
                                      FLAGS_ERROR_UI_FILTER_FOR_ERRORS       |
                                      FLAGS_ERROR_UI_FLAGS_GENERATE_DATA     |
                                      FLAGS_ERROR_UI_FLAGS_CHANGE_OPTIONS,
                                      NULL) != ERROR_SUCCESS ) {
                    MYDEBUGERR("InternetErrorDlg", GetLastError());
                    goto ERR_EXIT;
                }
            } else {
                goto ERR_EXIT;
            }
        }
    }

    ref(); // balanced by deref in onRequestComplete
    RHDEBUGMSG("will start onRequestComplete refCount = %d\n", refCount());

    if (d->m_loadSynchronously) {
        RHDEBUGMSG("call onRequestComplete with Synchronous\n");
        while (onRequestComplete() == RETURN_STATE_OK) {
            // Loop until finished.
        }
    } else {
        if (sts) return onRequestComplete();
        RHDEBUGMSG("ERROR_IO_PENDING HttpOpenRequestW\n");
    }
    return RETURN_STATE_OK;
ERR_EXIT:
    clearHandle();
    RHDEBUGMSG("start didFail\n");
    if (ResourceHandleClient* resourceHandleClient = client())
        resourceHandleClient->didFail(this, ResourceError(firstRequest().url().host(), dwError, firstRequest().url().string(), ""));
    return RETURN_STATE_NG;
}

void ResourceHandle::onRedirect() {
    ResourceRequest newRequest = firstRequest();
	//Multiple language at Redirect URL Fixed by Lupin NECSTHZ
    newRequest.setURL(KURL(ParsedURLString, d->m_redirectUrl,true));

    ResourceResponse response(firstRequest().url(), String(), 0, String(), String());

    if (ResourceHandleClient* resourceHandleClient = client())
        resourceHandleClient->willSendRequest(this, newRequest, response);
}

void ResourceHandle::onReadFile() {
    RHDEBUGMSG("start ThreadID = 0x%x, m_dwErrorCode = 0x%x, d->m_Overlapped.Offset = %d, m_dwNumberOfBytesTransfered = %d\n", GetCurrentThreadId(), d->m_dwErrorCode, d->m_Overlapped.Offset, d->m_dwNumberOfBytesTransfered);
    if (d->m_dwNumberOfBytesTransfered > 0) {
        if (ResourceHandleClient* resourceHandleClient = client())
            resourceHandleClient->didReceiveData(this, d->m_frbuffer, d->m_dwNumberOfBytesTransfered, -1);
        if (d->m_fileHandle && d->m_dwErrorCode != ERROR_HANDLE_EOF) {
            d->m_Overlapped.Offset += d->m_dwNumberOfBytesTransfered;
            d->m_Overlapped.OffsetHigh = 0;
            d->m_Overlapped.hEvent = (HANDLE)reinterpret_cast<DWORD_PTR>(this);
            RHDEBUGMSG("continue to ReadFileEx m_Overlapped.Offset = %d\n", d->m_Overlapped.Offset);
            if (ReadFileEx(d->m_fileHandle, &(d->m_frbuffer), RFBUFF_SIZE, &(d->m_Overlapped), &ResourceHandle::FileReadIOCompletionRoutine) == TRUE) {
                SleepEx( INFINITE, TRUE );
                //setMainThreadCallbacksPaused(false);
                return;
            }
        }
    }
    if (d->m_fileHandle) {
        CloseHandle(d->m_fileHandle);
        d->m_fileHandle = 0;
    }
    if (!d->m_canceled) {
        if (ResourceHandleClient* resourceHandleClient = client())
            resourceHandleClient->didFinishLoading(this, 0);
    }
    RHDEBUGMSG("start deref refCount = %d\n", refCount());
    deref(); // balances ref in start
    RHDEBUGMSG("end deref\n");
}


void ResourceHandle::uploadFormDataRoutine( void* argument ) {
    const static int uploadDataMaxSize   = 1024 * 16;      // 16K
    const static int readFileDataMaxSize = 1024 * 16;      // 16K

	RHDEBUGMSG("start ThreadID = 0x%x, ResourceHandle = 0x%x", GetCurrentThreadId(), argument);
	ResourceHandle *handle = (ResourceHandle *) argument;
	if (handle == NULL) {
		return;
	}

	char* buffers = new char[uploadDataMaxSize];
	bool canReadNext = true;
	size_t index = 0;
	Vector<char> data;
	const FormDataElement * e;
	bool continueRead = true;
	PlatformFileHandle hFile = NULL;
	bool isWaitForWriteFile = false;

	size_t n = handle->firstRequest().httpBody()->elements().size();
	DWORD writeFileTimeout = WebCore::httpDefaultTimeOut();
	DWORD dwBufferLength = sizeof(writeFileTimeout);
    if (writeFileTimeout == 0) {
		InternetQueryOption(handle->d->m_requestHandle, 
			INTERNET_OPTION_DATA_SEND_TIMEOUT,
			&writeFileTimeout, 
			&dwBufferLength);
    }

	do{

		DWORD bytesWillWrite = 0;
		DWORD bytesWritten = 0;
		DWORD dwError = 0;

		FillMemory(buffers, uploadDataMaxSize, 0);

		if (handle->d->m_bytesRemainingToWrite >= uploadDataMaxSize){
			bytesWillWrite = uploadDataMaxSize;
		} else {
			bytesWillWrite = handle->d->m_bytesRemainingToWrite;
		}

		do {

			if (data.size() >= bytesWillWrite) {
				CopyMemory(buffers, data.data(), bytesWillWrite);
				data.remove(0, bytesWillWrite);
				break;
			}

			if (canReadNext) {
				e = &handle->firstRequest().httpBody()->elements()[index];
				canReadNext = false;
			}

            switch(e->m_type) {
            case FormDataElement::data: {
                    data.append(e->m_data.data(), static_cast<size_t>(e->m_data.size()));
                    index ++;
                    canReadNext = true;

					break;
				}
			case FormDataElement::encodedFile: {
					if (!hFile){
						hFile = openFile(e->m_filename, FileOpenMode::OpenForRead);
					}

					int nReadSize = -1;

					if (hFile != INVALID_HANDLE_VALUE){
						char * readBuffer = new char[readFileDataMaxSize];
						nReadSize = readFromFile(hFile, readBuffer, readFileDataMaxSize);
						if (nReadSize > 0){
							data.append(readBuffer, nReadSize);
						} else {
							closeFile(hFile);
							hFile = NULL;
							index ++;
							canReadNext = true;
						}

						if (readBuffer != NULL) {
							delete[] readBuffer;
							readBuffer = NULL;
						}
						
					} 

					if (hFile == INVALID_HANDLE_VALUE || nReadSize == -1) {
						MYDEBUGERR("readFromFile Failed error = 0x%x FileName = %s ", GetLastError(), e->m_filename);
						handle->d->m_dwErrorCode = dwError;
						callOnMainThread(callOnRequestComplete, handle);
						goto Clearup;
					}

					break;
				}
			default:
				break;
			}

		} while(true);


		RHDEBUGMSG("InternetWriteFile handle = 0x%x,  RemainingSize = %d \n", handle->d->m_requestHandle, handle->d->m_bytesRemainingToWrite);	
		if (!InternetWriteFile(handle->d->m_requestHandle, buffers, bytesWillWrite, &bytesWritten)) {
			dwError = GetLastError();
			if (dwError != ERROR_IO_PENDING) {
				MYDEBUGERR("InternetWriteFile didFail", dwError);
				handle->d->m_dwErrorCode = dwError;
				handle->d->m_bytesRemainingToWrite = 0;
				goto Clearup;
			}
		}

        handle->d->m_bytesRemainingToWrite -= bytesWillWrite;
        if ( (handle->d->m_bytesRemainingToWrite > 0)  && (dwError == ERROR_IO_PENDING)) {
            DWORD dwObject = WaitForMultipleObjects(2, handle->m_requestCompleteEvent, false,  writeFileTimeout);
            switch(dwObject) {
            case WAIT_OBJECT_0 :
                break;
            case WAIT_OBJECT_0 + 1 :
                goto Clearup;
                break;
            case WAIT_TIMEOUT:
                MYDEBUGERR("WaitForSingleObject Timeout: %d", writeFileTimeout);
                goto Clearup;
                break;
            default:
                goto Clearup;
                break;
            }
        }

		if ( handle->d->m_bytesRemainingToWrite == 0) {
			if (dwError != ERROR_IO_PENDING) {
				callOnMainThread(callOnRequestComplete, handle);
			}

			WaitForSingleObject(handle->m_requestCompleteEvent[1], writeFileTimeout);
		}

	} while (handle->d->m_bytesRemainingToWrite > 0);

Clearup:

	if (hFile) {
		closeFile(hFile);
		hFile = NULL;
	}

	if (buffers) {
		delete[] buffers;
		buffers = NULL;
	}

	data.clear();

}

bool ResourceHandle::onRequestComplete() {
    RHDEBUGMSG("start this = 0x%x, client() = 0x%x, ThreadID = 0x%x, url = %s\n", this, client(), GetCurrentThreadId(), firstRequest().url().string().latin1().data());
#ifndef GLOBAL_INETHANDLE
    if (!d->m_internetHandle) { // 0 if canceled.
        deref(); // balances ref in start
        return false;
    }
#endif
    if (!d->m_requestHandle) {  // 0 if canceled.
        RHDEBUGMSG("Already canceled m_requestHandle = 0\n");
        goto ERR_EXIT;
    }

    if (d->m_dwErrorCode) {
        if (ResourceHandleClient* rhc = client()) {
            MYDEBUGMSG("start didFail m_requestHandle = 0x%x, m_dwErrorCode = %d\n", d->m_requestHandle, d->m_dwErrorCode);
            MYDEBUGMSG("url = %s\n", firstRequest().url().string().latin1().data());
            rhc->didFail(this, ResourceError(firstRequest().url().host(), d->m_dwErrorCode, firstRequest().url().string(), ""));
            RHDEBUGMSG("finish didFail\n");
        }

		if ( m_isCreateUploadFormDataThread ) {
			// notify the upload thread to close
			SetEvent(m_requestCompleteEvent[1]);
		}

        goto ERR_EXIT;
    }

//    if (d->m_bytesRemainingToWrite) {
//RETRY_WRITE:
//        DWORD bytesWritten;
//        DWORD dwError = 0;
//		DWORD dwWriteSize = d->m_bytesRemainingToWrite;
//        MYDEBUGMSG("InternetWriteFile m_uploadcnt = %d\n", d->m_uploadcnt);
//		d->m_uploadcnt--;
//		if (d->m_uploadcnt) {
//			dwWriteSize = 1024;
//		}
//        if (InternetWriteFile(d->m_requestHandle,
//                              d->m_formData.data() + (d->m_formData.size() - d->m_bytesRemainingToWrite),
//                              dwWriteSize,
//                              &bytesWritten) == FALSE) {
//            dwError = GetLastError();
//            if (dwError != ERROR_IO_PENDING) {
//                MYDEBUGERR("InternetWriteFile didFail", dwError);
//                if (ResourceHandleClient* resourceHandleClient = client())
//                    resourceHandleClient->didFail(this, ResourceError(firstRequest().url().host(), dwError, firstRequest().url().string(), ""));
//                goto ERR_EXIT;
//            }
//        }
//        d->m_bytesRemainingToWrite -= dwWriteSize;	// WebKit.org original code is "-= bytesWritten", but its not good. cos bytesWritten is zero in ERROR_IO_PENDING.
//        if (d->m_bytesRemainingToWrite > 0) {
//			if (d->m_uploadcnt) {
//				if (dwError == ERROR_IO_PENDING) {
//			        MYDEBUGMSG("InternetWriteFile ERROR_IO_PENDING bytesWritten = %d, m_uploadcnt = %d\n", bytesWritten, d->m_uploadcnt);
//					return RETURN_STATE_OK;
//				}
//		        MYDEBUGMSG("InternetWriteFile Fisnih_IO bytesWritten = %d, m_uploadcnt = %d\n", bytesWritten, d->m_uploadcnt);
//				goto RETRY_WRITE;
//			}
//            return RETURN_STATE_OK;
//        }
//        d->m_formData.clear();
//        if (dwError == ERROR_IO_PENDING) {
//            return RETURN_STATE_OK;
//        }
//    }

	if (d->m_hasRemainingToWrite) {
		if (d->m_bytesRemainingToWrite > 0){
			if (m_isCreateUploadFormDataThread) {
				SetEvent(m_requestCompleteEvent[0]);
			}else{
				createThread(uploadFormDataRoutine, this, "uploadFormDataRoutine");
				m_isCreateUploadFormDataThread = true;
			}
			return RETURN_STATE_OK;
		} else {
		   SetEvent(m_requestCompleteEvent[1]);
		   d->m_hasRemainingToWrite = false;
		}
	}

    if (!d->m_sentEndRequest) {
        d->m_sentEndRequest = true;
        RHDEBUGMSG("HttpEndRequestW\n");
        // download Ή HSR_DOWNLOAD g?
        if (!HttpEndRequestW(d->m_requestHandle, 0, HSR_INITIATE, d->m_loadSynchronously ? 0 : reinterpret_cast<DWORD_PTR>(this))) {
            DWORD dwError = GetLastError();
            if (dwError == ERROR_IO_PENDING) {
                RHDEBUGMSG("ERROR_IO_PENDING\n");
                return RETURN_STATE_OK;
            } else {
                MYDEBUGERR("HttpEndRequestW didFail", dwError);
                if (ResourceHandleClient* resourceHandleClient = client())
                    resourceHandleClient->didFail(this, ResourceError(firstRequest().url().host(), dwError, firstRequest().url().string(), ""));
                goto ERR_EXIT;
            }
        }
    }

#ifdef RHDEBUG
    char debugBuff[50];
#endif
    char buffer[INTERNET_BUFFER_SIZE];
    INTERNET_BUFFERSA buffers;
    do {
        FillMemory(&buffers, sizeof(INTERNET_BUFFERSA), 0);
        buffers.dwStructSize = sizeof(INTERNET_BUFFERSA);
        buffers.lpvBuffer = buffer;
        buffers.dwBufferLength = INTERNET_BUFFER_SIZE;
#ifdef RHDEBUG
        strcpy(buffer, "(no data)");
#endif
        if (d->m_requestHandle == NULL) break;
        RHDEBUGMSG("InternetReadFileExA start url = %s\n", firstRequest().url().string().latin1().data());
        BOOL bSts = InternetReadFileExA(d->m_requestHandle, &buffers, d->m_loadSynchronously ? 0 : (IRF_NO_WAIT | IRF_USE_CONTEXT | IRF_ASYNC), d->m_loadSynchronously ? 0 : reinterpret_cast<DWORD_PTR>(this));
        if (!bSts) {
            DWORD dwError = GetLastError();
            if (dwError == ERROR_IO_PENDING) {
                RHDEBUGMSG("ERROR_IO_PENDING\n");
                return RETURN_STATE_OK;
            } else {
                MYDEBUGERR("InternetReadFileExA didFail", dwError);
                if (ResourceHandleClient* resourceHandleClient = client())
                    resourceHandleClient->didFail(this, ResourceError(firstRequest().url().host(), dwError, firstRequest().url().string(), ""));
                goto ERR_EXIT;
            }
        }
        RHDEBUGMSG("InternetReadFileExA end buffers.dwBufferLength = %d\n", buffers.dwBufferLength);
#ifdef RHDEBUG
        if (buffers.dwBufferLength > 0) {
            FillMemory(debugBuff, 50, 0);
            strncpy(debugBuff, buffer, 49);
            RHDEBUGMSG("ReadData = %s\n", debugBuff);
        }
#endif
        if (buffers.dwBufferLength == INTERNET_BUFFER_SIZE && d->m_requestHandle == NULL) break;

        if (!d->m_hasReceivedResponse) {
            d->m_hasReceivedResponse = true;

            ResourceResponse response;
            if (d->m_redirectUrl.isEmpty()) {
                response.setURL(firstRequest().url());
            } else {
                // Flash Player sometime requires new URL.
                response.setURL(KURL(ParsedURLString, d->m_redirectUrl, true));
            }

            String httpStatusText = queryHTTPHeader(d->m_requestHandle, HTTP_QUERY_STATUS_TEXT);
            if (!httpStatusText.isNull())
                response.setHTTPStatusText(httpStatusText);

            String httpStatusCode = queryHTTPHeader(d->m_requestHandle, HTTP_QUERY_STATUS_CODE);
            if (!httpStatusCode.isNull())
                response.setHTTPStatusCode(httpStatusCode.toInt());

            String httpContentLength = queryHTTPHeader(d->m_requestHandle, HTTP_QUERY_CONTENT_LENGTH);
            if (!httpContentLength.isNull())
                response.setExpectedContentLength(httpContentLength.toInt());

            String httpContentType = queryHTTPHeader(d->m_requestHandle, HTTP_QUERY_CONTENT_TYPE);
            if (!httpContentType.isNull()) {
                response.setMimeType(extractMIMETypeFromMediaType(httpContentType));
                response.setTextEncodingName(extractCharsetFromMediaType(httpContentType));
            }
            if (WebCore::gzipContentEncoding() & 1) {
				String httpContentEncoding = queryHTTPHeader(d->m_requestHandle, HTTP_QUERY_CONTENT_ENCODING);
				if (!httpContentEncoding.isNull() && equalIgnoringCase(httpContentEncoding, "gzip")) {
	                MYDEBUGMSG("ContentEncoding is gzip!!\n");
                    if (!(d->m_gzipInflateBuf = (char*)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, INTERNET_BUFFER_SIZE))) {
						if (ResourceHandleClient* resourceHandleClient = client())
							resourceHandleClient->didFail(this, ResourceError(firstRequest().url().host(), ERR_CODE_GZIP_OPEN, firstRequest().url().string(), ""));
						goto ERR_EXIT;
					}
					/* zlib */
					memset(&d->m_zst, 0, sizeof(z_stream));
					if (inflateInit2(&d->m_zst, 16 + MAX_WBITS) != Z_OK) {
						HeapFree(GetProcessHeap(), 0, d->m_gzipInflateBuf);
						d->m_gzipInflateBuf = NULL;
						MYDEBUGMSG("Error !! gzip inflate. %s\n", (d->m_zst.msg) ? d->m_zst.msg : "???");
						if (ResourceHandleClient* resourceHandleClient = client())
							resourceHandleClient->didFail(this, ResourceError(firstRequest().url().host(), ERR_CODE_GZIP_OPEN, firstRequest().url().string(), (d->m_zst.msg) ? d->m_zst.msg : "???"));
						goto ERR_EXIT;
					}
					d->m_zst.next_out = (Bytef*)d->m_gzipInflateBuf;		/* o̓|C^ */
                    d->m_zst.avail_out = INTERNET_BUFFER_SIZE;    /* o̓obt@c */
				}
			}

            String httpHeaders = queryHTTPHeader(d->m_requestHandle, HTTP_QUERY_RAW_HEADERS_CRLF);
            if (!httpHeaders.isNull()) {
                int linetopPos = 0;
                int crlfPos = httpHeaders.find("\r\n", linetopPos);
                RHDEBUGMSG("response httpHeaders\n%s\n", httpHeaders.latin1().data());
                while (crlfPos > linetopPos + 2) {
                    int splitPos = httpHeaders.find(":", linetopPos + 1);
                    if (splitPos > linetopPos && splitPos < crlfPos) {
                        String strKey = httpHeaders.substring(linetopPos, splitPos - linetopPos).stripWhiteSpace();
                        String strValue = httpHeaders.substring(splitPos + 1, crlfPos - (splitPos + 1)).stripWhiteSpace();
						if (!d->m_gzipInflateBuf || !strKey.startsWith("Content-Encoding")) {	// gzip𓀍ς݂̏ꍇAwb_珜
							response.setHTTPHeaderField(strKey, strValue);
						}
                    }
                    linetopPos = crlfPos + 2;
                    crlfPos = httpHeaders.find("\r\n", linetopPos);
                }
            }
            if (ResourceHandleClient* resourceHandleClient = client()) {
                RHDEBUGMSG("didReceiveResponse start this = 0x%x, client() = 0x%x\n", this, resourceHandleClient);
                resourceHandleClient->didReceiveResponse(this, response);
            }
        }

		// When HTTP Status Code 304(Not Modified) is returned, dwBufferLength is zero and should call didReceiveResponse without didReceiveData
		if (buffers.dwBufferLength == 0) break;

        // FIXME: https://bugs.webkit.org/show_bug.cgi?id=19793
        // -1 means we do not provide any data about transfer size to inspector so it would use
        // Content-Length headers or content size to show transfer size.
        if (ResourceHandleClient* resourceHandleClient = client()) {
            RHDEBUGMSG("didReceiveData start this = 0x%x, client() = 0x%x\n", this, resourceHandleClient);
			if (d->m_gzipInflateBuf) {
				int status = Z_OK;
				d->m_zst.next_in = (Bytef*)buffer; /* ̓|C^ɖ߂ */
				d->m_zst.avail_in = buffers.dwBufferLength;
				while (status != Z_STREAM_END) {
		            if (d->m_zst.avail_in == 0) break; /* ͎cʂ[ɂȂ΍ēǂݍ */
					status = inflate(&d->m_zst, Z_NO_FLUSH); /* WJ */
					if (status != Z_OK && status != Z_STREAM_END) { /* G[ */
						MYDEBUGMSG("Error !! gzip inflate. %s\n", (d->m_zst.msg) ? d->m_zst.msg : "???");
						if (ResourceHandleClient* resourceHandleClient = client())
							resourceHandleClient->didFail(this, ResourceError(firstRequest().url().host(), ERR_CODE_GZIP_INFLATE, firstRequest().url().string(), (d->m_zst.msg) ? d->m_zst.msg : "???"));
						goto ERR_EXIT;
					}
					if (status == Z_STREAM_END || d->m_zst.avail_out == 0) { /*  ܂ o̓obt@s*/
#ifdef TRYCATCH
			            try {
#endif
                            resourceHandleClient->didReceiveData(this, d->m_gzipInflateBuf, INTERNET_BUFFER_SIZE - d->m_zst.avail_out, -1);
#ifdef TRYCATCH
			            } catch(...) {  // OLb`
            			    MYDEBUGMSG("didReceiveData exception!!!\n");
                			goto ERR_EXIT;
            			}
#endif
						if (status == Z_STREAM_END) {
							buffers.dwBufferLength = 0; // gzipInternetReadIH
							break;
						}
						d->m_zst.next_out = (Bytef*)d->m_gzipInflateBuf; /* o̓|C^ɖ߂ */
                        d->m_zst.avail_out = INTERNET_BUFFER_SIZE; /* o̓obt@cʂɖ߂ */
					}
				}
			} else {
#ifdef TRYCATCH
            try {
#endif
                resourceHandleClient->didReceiveData(this, buffer, buffers.dwBufferLength, -1);
#ifdef TRYCATCH
            } catch(...) {  // OLb`
                MYDEBUGMSG("didReceiveData exception!!!\n");
                goto ERR_EXIT;
            }
#endif
			}
            RHDEBUGMSG("didReceiveData end\n");
        }
    } while (buffers.dwBufferLength > 0);

    clearHandle();
    if (!d->m_canceled) {
        if (ResourceHandleClient* resourceHandleClient = client()) {
            RHDEBUGMSG("didFinishLoading start url = %s\n", firstRequest().url().string().latin1().data());
#ifdef TRYCATCH
            try {
#endif
	            resourceHandleClient->didFinishLoading(this, 0);
#ifdef TRYCATCH
            } catch(...) {  // OLb`
                MYDEBUGMSG("didFinishLoading exception!!!\n");
                goto ERR_EXIT;
            }
#endif
            RHDEBUGMSG("didFinishLoading end\n");
        }
    }
ERR_EXIT:
    clearHandle();
    RHDEBUGMSG("start deref refCount = %d\n", refCount());
    deref(); // balances ref in start
    RHDEBUGMSG("end deref\n");
    return RETURN_STATE_NG;
}

bool ResourceHandle::start(NetworkingContext* context) {
    if ((++g_thisCreateCounter) % g_iniMemCheckInterval == 0) {
		MEMORYSTATUSEX msex = { sizeof(MEMORYSTATUSEX) };
		GlobalMemoryStatusEx( &msex );
	    DWORD dwMemSize = (DWORD)(msex.ullTotalVirtual - msex.ullAvailVirtual);

        MYDEBUGMSG("=== memory check ===  CreateCounter = %d, dwMemoryLoad = %d, VirtualBytes = %d KB\n", g_thisCreateCounter, msex.dwMemoryLoad, dwMemSize / 1024);
        if (msex.dwMemoryLoad >= g_iniMaxMemoryLoad || dwMemSize >= g_iniMaxWorkingSet) {
            WebCore::gcController().garbageCollectNow();
			MYDEBUGMSG("=== finish garbageCollectNow!!! ===  \n");
        }
    }
    d->m_dwErrorCode = ERROR_SUCCESS;
    RHDEBUGMSG("start this = 0x%x, client() = 0x%x, ThreadID = 0x%x, port = %d, firstRequest().url() = %s\n", this, client(), GetCurrentThreadId(), firstRequest().url().port(), firstRequest().url().string().latin1().data());
    if (firstRequest().url().isLocalFile() || firstRequest().url().protocolIsData()) {
        ref(); // balanced by deref in fileLoadTimer
        if (d->m_loadSynchronously) {
            RHDEBUGMSG("start fileLoadTimer with Sync\n");
            fileLoadTimer(0);
            RHDEBUGMSG("end fileLoadTimer with Sync refCount = %d\n", refCount());
        } else {
            RHDEBUGMSG("start fileLoadTimer with Async\n");
            //d->m_fileLoadTimer.startOneShot(0.0);
            d->m_fileLoadTimer.startOneShot(0.01);  // as same as cUrl (10ms)
            // WebCore::gcController().garbageCollectNow();
            RHDEBUGMSG("end fileLoadTimer with Async\n");
        }
        return RETURN_STATE_OK;
    }

	HINTERNET tmp_internetHandle;
#ifdef GLOBAL_INETHANDLE
    if (d->m_loadSynchronously) {
        if (!g_internetHandleSync) {
            g_internetHandleSync = setInternetHandle(context->userAgent(), false);
            if (!g_internetHandleSync) return RETURN_STATE_NG;
        }
		tmp_internetHandle = g_internetHandleSync;
    } else {
        if (!g_internetHandle) {
            g_internetHandle = setInternetHandle(context->userAgent(), true);
            if (!g_internetHandle) return RETURN_STATE_NG;
        }
		tmp_internetHandle = g_internetHandle;
    }
#else
    RHDEBUGMSG("start internet access m_internetHandle = 0x%x\n", d->m_internetHandle);
    if (!d->m_internetHandle)
        d->m_internetHandle = setInternetHandle(context->userAgent(), d->m_loadSynchronously ? false : true);
    if (!d->m_internetHandle)
        return RETURN_STATE_NG;
	tmp_internetHandle = d->m_internetHandle;
#endif
	if (!firstRequest().httpUserAgent().isNull()) {
		RHDEBUGMSG("userAgent = %s\n", firstRequest().httpUserAgent().latin1().data());
		InternetSetOption(tmp_internetHandle, INTERNET_OPTION_USER_AGENT, (LPVOID)firstRequest().httpUserAgent().charactersWithNullTermination(), firstRequest().httpUserAgent().length());
	}
    d->m_flags = getInetFlags(firstRequest());
    RHDEBUGMSG("start InternetConnectW internetHandle = 0x%x, port = %d, host = %s\n", tmp_internetHandle, firstRequest().url().port(), firstRequest().url().host().latin1().data());
    d->m_connectHandle = InternetConnectW(tmp_internetHandle, firstRequest().url().host().charactersWithNullTermination(),
                                          firstRequest().url().port(), 0, 0, INTERNET_SERVICE_HTTP, d->m_flags, d->m_loadSynchronously ? 0 : reinterpret_cast<DWORD_PTR>(this));
    RHDEBUGMSG("finish InternetConnectW m_connectHandle = 0x%x\n", d->m_connectHandle);
    if (d->m_connectHandle) {
        return onRequestStart();
    } else {
        DWORD dwError = GetLastError();
        if (dwError != ERROR_IO_PENDING) {
            MYDEBUGERR("InternetConnectW didFail", dwError);
            if (ResourceHandleClient* resourceHandleClient = client())
                resourceHandleClient->didFail(this, ResourceError(firstRequest().url().host(), dwError, firstRequest().url().string(), ""));
            return RETURN_STATE_NG;
        }
    }
    RHDEBUGMSG("ERROR_IO_PENDING InternetConnectW\n");
    d->m_iowaiting = true;
    return RETURN_STATE_OK;
}

void ResourceHandle::fileLoadTimer(Timer<ResourceHandle>*) {
    RHDEBUGMSG("start ThreadID = 0x%x\n", GetCurrentThreadId());
    if (firstRequest().url().protocolIsData()) {
        handleDataURL(this);
        deref(); // balances ref in start
        return;
    }

    String fileName = firstRequest().url().fileSystemPath();
    d->m_fileHandle = CreateFileW(fileName.charactersWithNullTermination(), GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING, d->m_loadSynchronously ? FILE_ATTRIBUTE_NORMAL : FILE_FLAG_OVERLAPPED, 0);

    if (d->m_fileHandle == INVALID_HANDLE_VALUE) {
        DWORD dwError = GetLastError();
        MYDEBUGERR("CreateFileW didFail INVALID_HANDLE_VALUE!!", dwError);
        MYDEBUGMSG("CreateFile Name = %s\n", fileName.charactersWithNullTermination());
        if (ResourceHandleClient* resourceHandleClient = client())
            resourceHandleClient->didFail(this, ResourceError(fileName, dwError, fileName, ""));
        deref(); // balances ref in start
        return;
    }

    ResourceResponse response;

    response.setURL(firstRequest().url());
    int dotPos = fileName.reverseFind('.');
    int slashPos = fileName.reverseFind('\\');
    if (slashPos == notFound) slashPos = fileName.reverseFind('/');

    if (slashPos < dotPos && dotPos != -1) {
        String ext = fileName.substring(dotPos + 1);
        response.setMimeType(MIMETypeRegistry::getMIMETypeForExtension(ext));
    }

    if (ResourceHandleClient* resourceHandleClient = client())
        resourceHandleClient->didReceiveResponse(this, response);

    bool result = false;
    if (d->m_loadSynchronously) {
        DWORD bytesRead = 0;
        RHDEBUGMSG("start Synchronous Read\n");
        do {
            bytesRead = 0;
            result = ReadFile(d->m_fileHandle, &(d->m_frbuffer), RFBUFF_SIZE, &bytesRead, 0);
            // FIXME: https://bugs.webkit.org/show_bug.cgi?id=19793
            // -1 means we do not provide any data about transfer size to inspector so it would use
            // Content-Length headers or content size to show transfer size.
            if (result && bytesRead)
                if (ResourceHandleClient* resourceHandleClient = client())
                    resourceHandleClient->didReceiveData(this, d->m_frbuffer, bytesRead, -1);
            // Check for end of file.
        } while (result && bytesRead);
    } else {
        d->m_Overlapped.Internal = 0;
        d->m_Overlapped.InternalHigh = 0;
        d->m_Overlapped.Offset = 0;
        d->m_Overlapped.OffsetHigh = 0;
        d->m_Overlapped.hEvent = (HANDLE)reinterpret_cast<DWORD_PTR>(this);
        result = ReadFileEx(d->m_fileHandle, &(d->m_frbuffer), RFBUFF_SIZE, &(d->m_Overlapped), &ResourceHandle::FileReadIOCompletionRoutine);
        RHDEBUGMSG("ReadFileEx result = 0x%x\n", result);
        if (result == true) {
            SleepEx( INFINITE, TRUE );
            RHDEBUGMSG("finish SleepEx\n");
            return;
        }
    }
    CloseHandle(d->m_fileHandle);
    d->m_fileHandle = 0;
    if (!d->m_canceled) {
        if (ResourceHandleClient* resourceHandleClient = client())
            resourceHandleClient->didFinishLoading(this, 0);
    }
    RHDEBUGMSG("before deref refCount = %d\n", refCount());
    deref(); // balances ref in start
}

void ResourceHandle::cancel() {
    // RHDEBUGMSG("start m_requestHandle = 0x%x, m_loadSynchronously = %d, url = %s\n", d->m_requestHandle, d->m_loadSynchronously, firstRequest().url().string().latin1().data());
    d->m_canceled = true;
    if (d->m_requestHandle) {
        clearHandle();
    } else {
        if (d->m_fileHandle) {
            CloseHandle(d->m_fileHandle);
            d->m_fileHandle = 0;
        }
        if (d->m_fileLoadTimer.isActive()) {
            d->m_fileLoadTimer.stop();
            deref(); // balances ref in start
        }
    }
}

void ResourceHandle::loadResourceSynchronously(NetworkingContext* context, const ResourceRequest& request, StoredCredentials storedCredentials, ResourceError& error, ResourceResponse& response, Vector<char>& data) {
    UNUSED_PARAM(storedCredentials);
    RHDEBUGMSG("start ThreadId = 0x%x, Method = %s, url = %s\n", GetCurrentThreadId(), request.httpMethod().latin1().data(), request.url().string().latin1().data());

	bool bErrorSync = true;
	char reqHeadersFileName[256] = {0};
	char bodyFileName[256] = {0};
	char resHeadersFileName[256] = {0};
    char postFileName[256] = {0};

    if (g_ExternalExeFilePtr && (!request.url().isLocalFile() && !request.url().protocolIsData())) {
        bool result = false;
        STARTUPINFOA si = { sizeof(STARTUPINFOA) };
        si.dwFlags = STARTF_USESHOWWINDOW;
        si.wShowWindow = SW_HIDE;
        PROCESS_INFORMATION pi = {};
        char frbuffer[RFBUFF_SIZE];

        char tempPathName[MAX_PATH];
        int iPathLen = ::GetTempPathA(MAX_PATH, tempPathName);
        char* p = tempPathName + (iPathLen - 1);
        if (*p != '\\') {
            *(p + 1) = '\\';
            *(p + 2) = 0;
        }

        sprintf_s(reqHeadersFileName, 255, "%sREQHEADERS%x.txt", tempPathName, GetCurrentThreadId());
        Vector<UChar> httpHeadersReq =  getRequestHeaders(request);
        int iSizeOfHeaders = httpHeadersReq.size();
        char* bufHeads = (char*)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, iSizeOfHeaders + 1);
        WideCharToMultiByte(CP_ACP, WC_NO_BEST_FIT_CHARS, String(httpHeadersReq.data(), iSizeOfHeaders).charactersWithNullTermination(), -1, bufHeads, iSizeOfHeaders, NULL, NULL);
		httpHeadersReq.clear();
        RHDEBUGMSG("start HttpSendRequestExW sjislen = %d, httpHeaders = \n----\n%s\n----\n", strlen(bufHeads), bufHeads);
        FILE *file;
        file = fopen(reqHeadersFileName, "wb");
        fwrite(bufHeads, sizeof(char), strlen(bufHeads), file);
        fclose(file);
        HeapFree(GetProcessHeap(), 0, bufHeads);

        DWORD dwPostSize = 0;
        if (request.httpBody()) {
			Vector<char> m_formData;
	        request.httpBody()->flatten(m_formData);
			dwPostSize = m_formData.size();
            if (dwPostSize > 0) {
                sprintf_s(postFileName, 255, "%sPOSTDATA%x.txt", tempPathName, GetCurrentThreadId());
                FILE *psfile;
                psfile = fopen(postFileName, "wb");
                fwrite(m_formData.data(), sizeof(char), dwPostSize, psfile);
                fclose(psfile);
#ifdef RHDEBUGMSG
			    ZeroMemory(frbuffer, sizeof(frbuffer));
				memcpy(frbuffer, m_formData.data(), dwPostSize);
		        RHDEBUGMSG("post data = %s\n", frbuffer);
#endif
            }
			m_formData.clear();
        }

        String urlStr = request.url().path();
        String urlQuery = request.url().query();
        if (!urlQuery.isEmpty()) {
            urlStr.append('?');
            urlStr.append(urlQuery);
        }

		sprintf_s(frbuffer, RFBUFF_SIZE, "%s \"%s\" \"%s\" 0x%x %s %d %s \"%s\"", g_ExternalExeFilePtr, reqHeadersFileName, request.httpUserAgent().latin1().data(), getInetFlags(request), request.url().host().latin1().data(), request.url().port(), urlStr.latin1().data(), (dwPostSize ? postFileName : ""));
        RHDEBUGMSG("start CreateProcessW %s\n", frbuffer);
        if (CreateProcessA(NULL, frbuffer, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi)) {
            RHDEBUGMSG("start WaitForSingleObject\n");
            DWORD dwSts = WaitForSingleObject(pi.hProcess, 30000);  // webkitinet.exȅő30b҂
            DWORD dwExitCode;
            if (GetExitCodeProcess(pi.hProcess, &dwExitCode)) {
                if (dwExitCode == STILL_ACTIVE) TerminateProcess(pi.hProcess, 0);
            }
            CloseHandle(pi.hProcess);
            RHDEBUGMSG("finish WaitForSingleObject dwSts = %d\n", dwSts);
			if (dwSts == WAIT_FAILED || dwSts == WAIT_TIMEOUT) {
				MYDEBUGMSG("failed WaitForSingleObject!! (%s)\n", (dwSts == WAIT_FAILED ? "WAIT_FAILED" : "WAIT_TIMEOUT"));
				goto SYNC_EXIT;
			}
			if (dwExitCode != 0) {
	            MYDEBUGMSG("Error GetExitCodeProcess!! dwExitCode = %d\n", dwExitCode);
				goto SYNC_EXIT;
			}
        } else {
            MYDEBUGMSG("failed CreateProcessW\n");
            goto SYNC_EXIT;
        }
        DeleteFileA(reqHeadersFileName);
		reqHeadersFileName[0] = 0;
		if (postFileName[0] != 0) {
	        DeleteFileA(postFileName);
			postFileName[0] = 0;
		}

        sprintf_s(resHeadersFileName, 255, "%sRESHEADERS%x.txt", tempPathName, pi.dwProcessId);
        sprintf_s(bodyFileName, 255, "%sBODY%x.dat", tempPathName, pi.dwProcessId);
        HANDLE headersFile = CreateFileA(resHeadersFileName, GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
        if (headersFile == INVALID_HANDLE_VALUE) {
            DWORD dwError = GetLastError();
            MYDEBUGERR("CreateFileW didFail INVALID_HANDLE_VALUE!!", dwError);
            MYDEBUGMSG("CreateFile Name = %s\n", resHeadersFileName);
            goto SYNC_EXIT;
        }
        DWORD bufferSize = GetFileSize(headersFile, NULL);
        RHDEBUGMSG("GetFileSize headersFile size = %d\n", bufferSize);
        char* buf2 = (char*)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, bufferSize);
        DWORD bytesRead = 0;
        result = ReadFile(headersFile, buf2, bufferSize, &bytesRead, 0);
        CloseHandle(headersFile);
        DeleteFileA(resHeadersFileName);
		resHeadersFileName[0] = 0;
        String httpHeaders = String(buf2, bytesRead);
        HeapFree(GetProcessHeap(), 0, buf2);

        if (!httpHeaders.isNull()) {
            int lineNum = 0;
            int linetopPos = 0;
            int crlfPos = httpHeaders.find("\r\n", linetopPos);
            RHDEBUGMSG("httpHeaders = %s\n", httpHeaders.latin1().data());
            while (crlfPos >= linetopPos) {
                ++lineNum;
                int iLen = crlfPos - linetopPos;
                if (iLen) {
                    String strHeaderLine = String(httpHeaders.substring(linetopPos, iLen).stripWhiteSpace());
                    RHDEBUGMSG("strHeaderLine = %s\n", strHeaderLine.latin1().data());
                    switch(lineNum) {
                    case 1:
                        response.setHTTPStatusText(strHeaderLine);
                        break;
                    case 2:
                        response.setHTTPStatusCode(strHeaderLine.toInt());
                        break;
                    case 3:
                        response.setExpectedContentLength(strHeaderLine.toInt());
                        break;
                    case 4:
                        response.setMimeType(extractMIMETypeFromMediaType(strHeaderLine));
                        response.setTextEncodingName(extractCharsetFromMediaType(strHeaderLine));
                        break;
                    default: {
                            int splitPos = strHeaderLine.find(":", 0);
                            if (splitPos) {
                                response.setHTTPHeaderField(strHeaderLine.substring(0, splitPos).stripWhiteSpace(), strHeaderLine.substring(splitPos + 1).stripWhiteSpace());
                            }
                        }
                    }
                }
                linetopPos = crlfPos + 2;
                crlfPos = httpHeaders.find("\r\n", linetopPos);
            }
        }
        response.setURL(request.url());

        HANDLE bodyFile = CreateFileA(bodyFileName, GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
        if (bodyFile == INVALID_HANDLE_VALUE) {
            DWORD dwError = GetLastError();
            RHDEBUGERR("CreateFileW bodyFile is INVALID_HANDLE_VALUE!!", dwError);
            RHDEBUGMSG("CreateFile Name = %s\n", bodyFile);
            // Some normal HTTP response has no body, for example, 304, so its not error
			bErrorSync = false;
			goto SYNC_EXIT;
        }
        do {
            bytesRead = 0;
		    ZeroMemory(frbuffer, sizeof(frbuffer));
            result = ReadFile(bodyFile, frbuffer, RFBUFF_SIZE, &bytesRead, 0);
            if (result && bytesRead) {
                RHDEBUGMSG("data.append bytesRead = %d, frbuffer = %s\n", bytesRead, frbuffer);
                data.append(frbuffer, bytesRead);
            }
        } while (result && bytesRead);
        CloseHandle(bodyFile);
		bErrorSync = false;
		goto SYNC_EXIT;
    } else {
        WebCoreSynchronousLoader syncLoader(error, response, data, request.httpUserAgent());
        ResourceHandle* handle = new ResourceHandle(request, &syncLoader, true, false);
        handle->setSynchronousInternetHandle(syncLoader.internetHandle());
        handle->start(context);
        return;
    }
SYNC_EXIT:
	if (resHeadersFileName[0] != 0) DeleteFileA(resHeadersFileName);
    if (bodyFileName[0] != 0) DeleteFileA(bodyFileName);
	if (reqHeadersFileName[0] != 0) DeleteFileA(reqHeadersFileName);
	if (postFileName[0] != 0) DeleteFileA(postFileName);
	if (bErrorSync) {
    	error = ResourceError(request.url().host(), ERR_CODE_SYNC, request.url().string(), "");
	}
}

void ResourceHandle::setSynchronousInternetHandle(HINTERNET internetHandle) {
    RHDEBUGMSG("start\n");
#ifdef GLOBAL_INETHANDLE
    if (g_internetHandleSync != internetHandle) {
        MYDEBUGMSG("********* Error !!! ********* g_internetHandle != internetHandle (0x%x != 0x%x)\n", g_internetHandle, internetHandle);
    }
    g_internetHandleSync = internetHandle;
#else
    d->m_internetHandle = internetHandle;
#endif
    d->m_loadSynchronously = true;
}

bool ResourceHandle::willLoadFromCache(ResourceRequest&, Frame*) {
    MYDEBUGMSG("start  **** notImplemented !!! ****\n");
    notImplemented();
    return false;
}

void prefetchDNS(const String&) {
    RHDEBUGMSG("start  **** notImplemented !!! ****\n");
    notImplemented();
}

bool ResourceHandle::loadsBlocked() {
    RHDEBUGMSG("start  **** notImplemented !!! ****\n");
    return false;
}

void ResourceHandle::platformSetDefersLoading(bool bDefer) {
    RHDEBUGMSG("start  **** notImplemented !!! **** this = 0x%x, bDefer = %d, url = %s\n", this, bDefer, firstRequest().url().string().latin1().data());
//  setMainThreadCallbacksPaused(bDefer);
    notImplemented();
}

} // namespace WebCore
